
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.26.15',
    'branch_nick': 'release-1.26-LTS',
    'revision_id': '5c3f075fd25c4986aa53869ee76218a841d94c40',
    'revno': 10949
}
